# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import Any


class DataAPIDictUDT(dict[Any, Any]):
    """
    A `dict` subclass acting as a marker for wrapping user-defined type (UDT) data.

    This object behaves entirely as a regular Python `dict`.
    Additionally, however, when encountered within a payload to be sent to the
    Data API, its ensures its contents are serialized as JSON payload in a
    format suitable for transmitting UDT data.

    Note that, depending on the Table object's `SerdesOptions`, other representations
    may be available for encoding UDT data (see the package README for more options):
    still, usage of `DataAPIDictUDT` is the one choice that works regardless of the
    serdes settings.

    When reading from Tables, correspondingly, if custom data types are enabled
    (as per default serdes options), UDT data is returned as `DataAPIDictUDT`
    dictionaries unless more specific deserialization rules are configured.
    """

    def __init__(
        self,
        initial: Mapping[str, Any] | Iterable[tuple[Any, Any]] | None = None,
        **kwargs: Any,
    ) -> None:
        if initial is None:
            initial = {}
        super().__init__(initial, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
from decimal import Decimal

import pytest
from httpx import HTTPStatusError, Response
from pytest_httpserver import HTTPServer

from astrapy import DataAPIClient
from astrapy.api_options import APIOptions, SerdesOptions
from astrapy.constants import Environment
from astrapy.exceptions import (
    DataAPIErrorDescriptor,
    DataAPIHttpException,
    DataAPIResponseException,
    InvalidEnvironmentException,
)

ERROR_MESSAGE = "da_message"
ERROR_CODE = "DA_ERRORCODE"
ERROR_ATTRIBUTES = {"ekey": "evalue"}
ERROR_TITLE = "da_title"
ERROR_JSON = {
    "title": ERROR_TITLE,
    "errorCode": ERROR_CODE,
    "message": ERROR_MESSAGE,
    **ERROR_ATTRIBUTES,
}
ERROR_TITLE = f"{ERROR_TITLE}: {ERROR_MESSAGE} ({ERROR_CODE})"
FULL_RESPONSE_OF_500 = json.dumps({"errors": [ERROR_JSON]})


@pytest.mark.describe("test DataAPIHttpException")
def test_dataapihttpexception() -> None:
    """Test that regardless of how incorrect the input httpx error, nothing breaks."""
    se0 = HTTPStatusError(message="httpx_message", request="req", response=None)  # type: ignore[arg-type]
    se1 = HTTPStatusError(message="httpx_message", request="req", response="blah")  # type: ignore[arg-type]
    se2 = HTTPStatusError(
        message="httpx_message",
        request="req",  # type: ignore[arg-type]
        response=Response(status_code=500, text="blah"),
    )
    se3 = HTTPStatusError(
        message="httpx_message",
        request="req",  # type: ignore[arg-type]
        response=Response(status_code=500, text='{"blabla": 1}'),
    )
    se4 = HTTPStatusError(
        message="httpx_message",
        request="req",  # type: ignore[arg-type]
        response=Response(status_code=500, text='{"errors": []}'),
    )
    se5 = HTTPStatusError(
        message="httpx_message",
        request="req",  # type: ignore[arg-type]
        response=Response(
            status_code=500,
            text=FULL_RESPONSE_OF_500,
        ),
    )

    de0 = DataAPIHttpException.from_httpx_error(se0)
    de1 = DataAPIHttpException.from_httpx_error(se1)
    de2 = DataAPIHttpException.from_httpx_error(se2)
    de3 = DataAPIHttpException.from_httpx_error(se3)
    de4 = DataAPIHttpException.from_httpx_error(se4)
    de5 = DataAPIHttpException.from_httpx_error(se5)

    repr(de0)
    repr(de1)
    repr(de2)
    repr(de3)
    repr(de4)
    repr(de5)
    str(de0)
    str(de1)
    str(de2)
    str(de3)
    str(de4)
    assert ERROR_MESSAGE in str(de5)


@pytest.mark.describe("test DataAPIHttpException raising 500 from a mock server, sync")
def test_dataapihttpexception_raising_500_sync(httpserver: HTTPServer) -> None:
    """
    testing that:
        - the request gets sent
        - the correct exception is raised, with the expected members:
            - X its request and response fields are those of the httpx object
            - its DataAPIHttpException fields are set
        - it is caught with an httpx "except" clause all right
    """
    root_endpoint = httpserver.url_for("/")
    client = DataAPIClient(environment="other")
    database = client.get_database(root_endpoint, keyspace="xkeyspace")
    collection = database.get_collection("xcoll")
    expected_url = "/v1/xkeyspace/xcoll"
    httpserver.expect_oneshot_request(
        expected_url,
        method="POST",
    ).respond_with_data(
        FULL_RESPONSE_OF_500,
        status=500,
    )
    exc = None
    try:
        collection.find_one()
    except HTTPStatusError as e:
        exc = e

    assert isinstance(exc, DataAPIHttpException)
    httpx_payload = json.loads(exc.request.content.decode())
    assert "findOne" in httpx_payload
    assert ERROR_MESSAGE in exc.response.text
    assert exc.error_descriptors == [DataAPIErrorDescriptor(ERROR_JSON)]
    assert ERROR_MESSAGE in str(exc)


@pytest.mark.describe("test DataAPIHttpException raising 404 from a mock server, sync")
def test_dataapihttpexception_raising_404_sync(httpserver: HTTPServer) -> None:
    """
    testing that:
        - the request gets sent
        - the correct exception is raised, with the expected members:
            - X its request and response fields are those of the httpx object
            - its DataAPIHttpException fields are set
        - it is caught with an httpx "except" clause all right
    """
    root_endpoint = httpserver.url_for("/")
    client = DataAPIClient(environment="other")
    database = client.get_database(root_endpoint, keyspace="xkeyspace")
    collection = database.get_collection("xcoll")
    expected_url = "/v1/xkeyspace/xcoll"
    httpserver.expect_oneshot_request(
        expected_url,
        method="POST",
    ).respond_with_data(
        "blah",
        status=404,
    )
    exc = None
    try:
        collection.find_one()
    except HTTPStatusError as e:
        exc = e

    assert isinstance(exc, DataAPIHttpException)
    httpx_payload = json.loads(exc.request.content.decode())
    assert "findOne" in httpx_payload
    assert "blah" in exc.response.text
    assert exc.error_descriptors == []
    # not parsable into a DataAPIErrorDescriptor -> not in the error str repr
    assert "blah" not in str(exc)


@pytest.mark.describe("test DataAPIHttpException raising 500 from a mock server, async")
async def test_dataapihttpexception_raising_500_async(httpserver: HTTPServer) -> None:
    """
    testing that:
        - the request gets sent
        - the correct exception is raised, with the expected members:
            - X its request and response fields are those of the httpx object
            - its DataAPIHttpException fields are set
        - it is caught with an httpx "except" clause all right
    """
    root_endpoint = httpserver.url_for("/")
    client = DataAPIClient(environment="other")
    adatabase = client.get_async_database(root_endpoint, keyspace="xkeyspace")
    acollection = adatabase.get_collection("xcoll")
    expected_url = "/v1/xkeyspace/xcoll"
    httpserver.expect_oneshot_request(
        expected_url,
        method="POST",
    ).respond_with_data(
        FULL_RESPONSE_OF_500,
        status=500,
    )
    exc = None
    try:
        await acollection.find_one()
    except HTTPStatusError as e:
        exc = e

    assert isinstance(exc, DataAPIHttpException)
    httpx_payload = json.loads(exc.request.content.decode())
    assert "findOne" in httpx_payload
    assert ERROR_MESSAGE in exc.response.text
    assert exc.error_descriptors == [DataAPIErrorDescriptor(ERROR_JSON)]
    assert ERROR_MESSAGE in str(exc)


@pytest.mark.describe("test DataAPIHttpException raising 404 from a mock server, async")
async def test_dataapihttpexception_raising_404_async(httpserver: HTTPServer) -> None:
    """
    testing that:
        - the request gets sent
        - the correct exception is raised, with the expected members:
            - X its request and response fields are those of the httpx object
            - its DataAPIHttpException fields are set
        - it is caught with an httpx "except" clause all right
    """
    root_endpoint = httpserver.url_for("/")
    client = DataAPIClient(environment="other")
    adatabase = client.get_async_database(root_endpoint, keyspace="xkeyspace")
    acollection = adatabase.get_collection("xcoll")
    expected_url = "/v1/xkeyspace/xcoll"
    httpserver.expect_oneshot_request(
        expected_url,
        method="POST",
    ).respond_with_data(
        "blah",
        status=404,
    )
    exc = None
    try:
        await acollection.find_one()
    except HTTPStatusError as e:
        exc = e

    assert isinstance(exc, DataAPIHttpException)
    httpx_payload = json.loads(exc.request.content.decode())
    assert "findOne" in httpx_payload
    assert "blah" in exc.response.text
    assert exc.error_descriptors == []
    # not parsable into a DataAPIErrorDescriptor -> not in the error str repr
    assert "blah" not in str(exc)


@pytest.mark.describe("test DataAPIResponseException")
def test_dataapiresponseexception() -> None:
    da_e1 = DataAPIResponseException.from_response(
        command={"cmd": "C1"},
        raw_response=({"errors": [ERROR_JSON]}),
    )
    the_daed = DataAPIErrorDescriptor(ERROR_JSON)

    assert the_daed.error_code == ERROR_CODE
    assert the_daed.message == ERROR_MESSAGE
    assert the_daed.attributes == ERROR_ATTRIBUTES

    assert da_e1.text == ERROR_TITLE
    assert str(da_e1) == ERROR_TITLE
    assert da_e1.error_descriptors == [the_daed]


@pytest.mark.describe("test of database info failures, sync")
def test_get_database_info_failures_sync() -> None:
    client = DataAPIClient(environment=Environment.OTHER)
    db = client.get_database("http://not.a.real.thing", keyspace="ks")
    with pytest.raises(InvalidEnvironmentException):
        db.info()


@pytest.mark.describe("test of database info failures, async")
async def test_get_database_info_failures_async() -> None:
    client = DataAPIClient(environment=Environment.OTHER)
    adb = client.get_async_database("http://not.a.real.thing", keyspace="ks")
    with pytest.raises(InvalidEnvironmentException):
        await adb.info()


@pytest.mark.describe("test of collections not inserting Decimals, sync")
def test_collections_error_on_decimal_sync(httpserver: HTTPServer) -> None:
    root_endpoint = httpserver.url_for("/")
    client = DataAPIClient(environment="other")
    database = client.get_database(root_endpoint, keyspace="xkeyspace")
    collection = database.get_collection("xcoll")
    expected_url = "/v1/xkeyspace/xcoll"
    httpserver.expect_request(
        expected_url,
        method="POST",
    ).respond_with_json({"status": {"insertedIds": ["x"]}})
    collection.insert_one({"a": 1.23})
    with pytest.raises(TypeError):
        collection.insert_one({"a": Decimal("1.23")})
    hd_collection = collection.with_options(
        api_options=APIOptions(
            serdes_options=SerdesOptions(use_decimals_in_collections=True)
        )
    )
    hd_collection.insert_one({"a": 1.23})
    hd_collection.insert_one({"a": Decimal("1.23")})


@pytest.mark.describe("test of collections not inserting Decimals, async")
async def test_collections_error_on_decimal_async(httpserver: HTTPServer) -> None:
    root_endpoint = httpserver.url_for("/")
    client = DataAPIClient(environment="other")
    adatabase = client.get_async_database(root_endpoint, keyspace="xkeyspace")
    acollection = adatabase.get_collection("xcoll")
    expected_url = "/v1/xkeyspace/xcoll"
    httpserver.expect_request(
        expected_url,
        method="POST",
    ).respond_with_json({"status": {"insertedIds": ["x"]}})
    await acollection.insert_one({"a": 1.23})
    with pytest.raises(TypeError):
        await acollection.insert_one({"a": Decimal("1.23")})
    hd_acollection = acollection.with_options(
        api_options=APIOptions(
            serdes_options=SerdesOptions(use_decimals_in_collections=True)
        )
    )
    await hd_acollection.insert_one({"a": 1.23})
    await hd_acollection.insert_one({"a": Decimal("1.23")})



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import argparse
import json
import os
from typing import Any, no_type_check


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Compare old and new coverage JSON data and print a report."
    )

    # Mandatory positional arguments
    parser.add_argument("old_coverage_json", help="Path to the old coverage JSON file")
    parser.add_argument("new_coverage_json", help="Path to the new coverage JSON file")

    # Optional output filename
    parser.add_argument(
        "-o",
        "--output",
        metavar="filename",
        help="Write output to a file instead of stdout",
    )

    # Optional format selector
    parser.add_argument(
        "-f",
        "--format",
        choices=["json", "table"],
        default="json",
        help="Output format (default: json)",
    )

    return parser.parse_args()


def calc_delta(prev_j0: dict[str, Any] | None, new_j: dict[str, Any]) -> dict[str, Any]:
    # Workaround to missing prev (we treat is as 'no previous coverage')
    has_delta = prev_j0 is not None
    prev_j = prev_j0 or {"totals": {"percent_covered": 0.0}, "files": {}}
    overall_delta = (
        new_j["totals"]["percent_covered"] - prev_j["totals"]["percent_covered"]
    )

    def _file_delta(pj: dict[str, Any], nj: dict[str, Any], f_name: str) -> float:
        new_coverage: float = nj["files"][f_name]["summary"]["percent_covered"]
        if f_name not in pj["files"]:
            return new_coverage
        else:
            old_coverage: float = pj["files"][f_name]["summary"]["percent_covered"]
            return new_coverage - old_coverage

    per_file_deltas: dict[str, Any] = dict(
        sorted(
            [
                (
                    new_file_k,
                    {
                        **(
                            {"delta": _file_delta(prev_j, new_j, f_name=new_file_k)}
                            if has_delta
                            else {}
                        ),
                        **new_file_v["summary"],
                    },
                )
                for new_file_k, new_file_v in new_j["files"].items()
            ],
            key=lambda file_pair: (
                file_pair[1].get("delta", 0.0),
                file_pair[1]["percent_covered"],
            ),
        )
    )
    return {
        "totals": {
            **({"delta": overall_delta} if has_delta else {}),
            **new_j["totals"],
        },
        "files": per_file_deltas,
    }


def make_json_report(r_json: dict[str, float]) -> str:
    return json.dumps(r_json, indent=2)


@no_type_check
def make_table_report(r_json: dict[str, float], print_delta: bool = True) -> str:
    TABLE_GUTTER = 3
    FMT_NUM_DECIMALS = 2
    FMT_STR = f"% {FMT_NUM_DECIMALS + 4}.{FMT_NUM_DECIMALS}f%%"
    COLUMNERS_0 = [
        ("File", lambda kp: kp[0]),
        ("Stmts", lambda kp: f"{(kp[1]['num_statements'])}"),
        ("Miss", lambda kp: f"{(kp[1]['missing_lines'])}"),
        ("Cover", lambda kp: FMT_STR % kp[1]["percent_covered"]),
    ]
    COLUMNERS_D = [
        ("Delta", lambda kp: FMT_STR % kp[1]["delta"]),
    ]
    COLUMNERS = COLUMNERS_0 + (COLUMNERS_D if print_delta else [])

    col_widths0 = [
        max(len(col_n), len(col_v(("totals", r_json["totals"]))))
        for col_n, col_v in COLUMNERS
    ]
    col_widths = [
        max(
            [len(col_v((file_k, file_v))) for file_k, file_v in r_json["files"].items()]
            + [col_head_max]
        )
        for (_, col_v), col_head_max in zip(COLUMNERS, col_widths0)
    ]

    def _mk_line(file_k, file_v):
        return (" " * TABLE_GUTTER).join(
            [
                (f"%{col_w}s") % col_v((file_k, file_v))
                for (_, col_v), col_w in zip(COLUMNERS, col_widths)
            ]
        )

    header_line = (" " * TABLE_GUTTER).join(
        [(f"%{col_w}s") % col_k for (col_k, _), col_w in zip(COLUMNERS, col_widths)]
    )
    totals_line = _mk_line("totals", r_json["totals"])
    line_width = sum(col_widths) + TABLE_GUTTER * (len(COLUMNERS) - 1)
    sep_line = "-" * line_width

    file_lines = [
        _mk_line(file_k, file_v) for file_k, file_v in r_json["files"].items()
    ]

    return "\n".join([header_line, sep_line] + file_lines + [sep_line, totals_line])


if __name__ == "__main__":
    args = parse_args()
    old_coverage: dict[str, Any] | None
    has_delta = os.path.isfile(args.old_coverage_json)
    if has_delta:
        old_coverage = json.load(open(args.old_coverage_json))
    else:
        old_coverage = None
    new_coverage = json.load(open(args.new_coverage_json))
    delta_json = calc_delta(old_coverage, new_coverage)
    report_str: str
    if args.format == "json":
        report_str = make_json_report(delta_json)
    elif args.format == "table":
        report_str = make_table_report(delta_json, print_delta=has_delta)
    else:
        raise ValueError(f"Unknown report format {args.format}.")

    if args.output is None:
        print(report_str)
    else:
        print(f"Writing report to {args.output} ...", end="")
        with open(args.output, "w") as ofile:
            ofile.write(report_str)
        print(" done.")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.cursors import RerankedResult
from astrapy.data_types import DataAPIVector

from ..conftest import IS_ASTRA_DB, USE_RERANKER_API_KEY_HEADER, DefaultAsyncCollection

RUN_RERANKER_TESTS = IS_ASTRA_DB or USE_RERANKER_API_KEY_HEADER


@pytest.mark.skipif(
    not RUN_RERANKER_TESTS,
    reason="Reranker-related tests skipped (no Astra and no header reranker key provided)",
)
class TestCollectionFindAndRerankAsync:
    @pytest.mark.describe("test of collection find-and-rerank vectorize, async")
    async def test_collection_farr_vectorize_async(
        self,
        async_empty_farr_vectorize_collection: DefaultAsyncCollection,
    ) -> None:
        acoll = async_empty_farr_vectorize_collection
        # insertions
        await acoll.insert_many(
            [
                {
                    "_id": "01",
                    "$vectorize": "this is a cat",
                    "$lexical": "a cat",
                    "tag": "test01",
                },
                {
                    "_id": "01b",
                    "$vectorize": "this is a lynx",
                    "$lexical": "a lynx",
                    "tag": "test01",
                },
                {
                    "_id": "02",
                    "$hybrid": "this is a dog",
                    "tag": "test01",
                },
                {
                    "_id": "03",
                    "$hybrid": {
                        "$vectorize": "this is a Pucciniomycotina",
                        "$lexical": "Pucciniomycotina, my dear rust fungus",
                    },
                    "tag": "test01",
                },
            ]
        )
        # find-and-rerank functional test
        farr_cursor = acoll.find_and_rerank(
            {},
            sort={"$hybrid": "bla"},
            projection={"$vectorize": True},
            include_scores=True,
            limit=5,
        )

        hits = await farr_cursor.to_list()
        assert len(hits) == 4
        assert all(isinstance(hit, RerankedResult) for hit in hits)
        assert all(isinstance(hit.document, dict) for hit in hits)
        assert all(isinstance(hit.document["$vectorize"], str) for hit in hits)
        assert all(isinstance(hit.scores, dict) for hit in hits)
        assert all(len(hit.scores) > 0 for hit in hits)
        # some scores can get back as None or integer
        assert all(
            all(isinstance(sc, (float, int, type(None))) for sc in hit.scores.values())
            for hit in hits
        )
        assert all(
            any(isinstance(sc, float) for sc in hit.scores.values()) for hit in hits
        )

        # sort.$hybrid can be an object as well:
        farr_cursor_s_o = acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vectorize": "bla", "$lexical": "bla"}},
            projection={"$vectorize": True},
            include_scores=True,
            limit=5,
        )
        s_o_hits = await farr_cursor_s_o.to_list()
        # scores may differ by epsilon
        assert [rr.document for rr in s_o_hits] == [rr.document for rr in hits]

        # hybrid_limits various forms, functional tests
        cur_no_hl = acoll.find_and_rerank(
            {},
            sort={"$hybrid": "bla"},
            limit=2,
        )
        cur_nu_hl = acoll.find_and_rerank(
            {},
            sort={"$hybrid": "bla"},
            limit=2,
            hybrid_limits=4,
        )
        cur_ob_hl = acoll.find_and_rerank(
            {},
            sort={"$hybrid": "bla"},
            limit=2,
            hybrid_limits={
                "$lexical": 4,
                "$vector": 3,
            },
        )
        assert len(await cur_no_hl.to_list()) == 2
        assert len(await cur_nu_hl.to_list()) == 2
        assert len(await cur_ob_hl.to_list()) == 2

    @pytest.mark.describe("test of collection find-and-rerank novectorize, async")
    async def test_collection_farr_novectorize_async(
        self,
        async_empty_farr_vector_collection: DefaultAsyncCollection,
    ) -> None:
        acoll = async_empty_farr_vector_collection
        # insertions
        await acoll.insert_many(
            [
                {
                    "_id": "01",
                    "text_content": "this is a cat",
                    "$vector": [1, 2],
                    "$lexical": "a cat",
                    "tag": "test01",
                },
                {
                    "_id": "01b",
                    "text_content": "this is a lynx",
                    "$vector": [2, 1],
                    "$lexical": "a lynx",
                    "tag": "test01",
                },
            ]
        )
        # find-and-rerank functional test
        farr_cursor = acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            projection={"$vector": True},
            include_scores=True,
            limit=5,
            rerank_on="text_content",
            rerank_query="blaa",
        )

        hits = await farr_cursor.to_list()
        assert len(hits) == 2
        assert all(isinstance(hit, RerankedResult) for hit in hits)
        assert all(isinstance(hit.document, dict) for hit in hits)
        assert all(isinstance(hit.document["$vector"], DataAPIVector) for hit in hits)
        assert all(len(hit.document["$vector"]) == 2 for hit in hits)
        assert all(isinstance(hit.scores, dict) for hit in hits)
        assert all(len(hit.scores) > 0 for hit in hits)
        # some scores can get back as None or integer
        assert all(
            all(isinstance(sc, (float, int, type(None))) for sc in hit.scores.values())
            for hit in hits
        )
        assert all(
            any(isinstance(sc, float) for sc in hit.scores.values()) for hit in hits
        )

        hits_dav = await acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": DataAPIVector([0, 1]), "$lexical": "bla"}},
            projection={"$vector": True},
            include_scores=True,
            limit=5,
            rerank_on="text_content",
            rerank_query="blaa",
        ).to_list()
        assert hits_dav == hits

        # hybrid_limits various forms, functional tests
        cur_no_hl = acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            limit=5,
            rerank_on="text_content",
            rerank_query="blaa",
        )
        cur_nu_hl = acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            limit=5,
            rerank_on="text_content",
            rerank_query="blaa",
            hybrid_limits=3,
        )
        cur_ob_hl = acoll.find_and_rerank(
            {},
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            limit=5,
            rerank_on="text_content",
            rerank_query="blaa",
            hybrid_limits={
                "$lexical": 4,
                "$vector": 3,
            },
        )
        assert len(await cur_no_hl.to_list()) == 2
        assert len(await cur_nu_hl.to_list()) == 2
        assert len(await cur_ob_hl.to_list()) == 2

    @pytest.mark.describe(
        "test of collection find-and-rerank include_scores, vectorize, async"
    )
    async def test_collection_includescores_farr_vectorize_async(
        self,
        async_empty_farr_vectorize_collection: DefaultAsyncCollection,
    ) -> None:
        acoll = async_empty_farr_vectorize_collection
        await acoll.insert_one({"$vectorize": "text", "$lexical": "text"})

        cur_n = acoll.find_and_rerank(sort={"$hybrid": "bla"})
        cur_f = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_scores=False,
        )
        cur_t = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_scores=True,
        )
        itm_n = await cur_n.__anext__()
        itm_f = await cur_f.__anext__()
        itm_t = await cur_t.__anext__()

        assert itm_n.scores == {}
        assert itm_f.scores == {}
        assert itm_t.scores != {}
        assert all(
            isinstance(val, (float, int, type(None))) for val in itm_t.scores.values()
        )

    @pytest.mark.describe(
        "test of collection find-and-rerank include_scores, novectorize, async"
    )
    async def test_collection_includescores_farr_novectorize_async(
        self,
        async_empty_farr_vector_collection: DefaultAsyncCollection,
    ) -> None:
        acoll = async_empty_farr_vector_collection
        await acoll.insert_one(
            {"$vector": [11, 12], "$lexical": "text", "content": "text"}
        )

        cur_n = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
        )
        cur_f = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_scores=False,
        )
        cur_t = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_scores=True,
        )
        itm_n = await cur_n.__anext__()
        itm_f = await cur_f.__anext__()
        itm_t = await cur_t.__anext__()

        assert itm_n.scores == {}
        assert itm_f.scores == {}
        assert itm_t.scores != {}
        assert all(
            isinstance(val, (float, int, type(None))) for val in itm_t.scores.values()
        )

    @pytest.mark.describe(
        "test of collection find-and-rerank get_sort_vector, vectorize, async"
    )
    async def test_collection_getsortvector_farr_vectorize_async(
        self,
        async_empty_farr_vectorize_collection: DefaultAsyncCollection,
    ) -> None:
        acoll = async_empty_farr_vectorize_collection
        await acoll.insert_one({"$vectorize": "text", "$lexical": "text"})

        cur_n0 = acoll.find_and_rerank(sort={"$hybrid": "bla"})
        cur_f0 = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_sort_vector=False,
        )
        cur_t0 = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_sort_vector=True,
        )

        assert await cur_n0.get_sort_vector() is None
        assert await cur_f0.get_sort_vector() is None
        gsv_t0 = await cur_t0.get_sort_vector()
        assert isinstance(gsv_t0, (list, DataAPIVector))
        assert isinstance(gsv_t0[0], float)

        cur_n1 = acoll.find_and_rerank(sort={"$hybrid": "bla"})
        cur_f1 = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_sort_vector=False,
        )
        cur_t1 = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_sort_vector=True,
        )
        await cur_n1.__anext__()
        await cur_f1.__anext__()
        await cur_t1.__anext__()

        assert await cur_n1.get_sort_vector() is None
        assert await cur_f1.get_sort_vector() is None
        gsv_t1 = await cur_t1.get_sort_vector()
        assert isinstance(gsv_t1, (list, DataAPIVector))
        assert isinstance(gsv_t1[0], float)

        cur_n2 = acoll.find_and_rerank(sort={"$hybrid": "bla"})
        cur_f2 = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_sort_vector=False,
        )
        cur_t2 = acoll.find_and_rerank(
            sort={"$hybrid": "bla"},
            include_sort_vector=True,
        )
        await cur_n2.to_list()
        await cur_f2.to_list()
        await cur_t2.to_list()

        assert await cur_n2.get_sort_vector() is None
        assert await cur_f2.get_sort_vector() is None
        gsv_t2 = await cur_t2.get_sort_vector()
        assert isinstance(gsv_t2, (list, DataAPIVector))
        assert isinstance(gsv_t2[0], float)

    @pytest.mark.describe(
        "test of collection find-and-rerank get_sort_vector, novectorize, async"
    )
    async def test_collection_getsortvector_farr_novectorize_async(
        self,
        async_empty_farr_vector_collection: DefaultAsyncCollection,
    ) -> None:
        acoll = async_empty_farr_vector_collection
        await acoll.insert_one(
            {"$vector": [11, 12], "$lexical": "text", "content": "text"}
        )

        cur_n0 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
        )
        cur_f0 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_sort_vector=False,
        )
        cur_t0 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_sort_vector=True,
        )

        assert await cur_n0.get_sort_vector() is None
        assert await cur_f0.get_sort_vector() is None
        gsv_t0 = await cur_t0.get_sort_vector()
        assert isinstance(gsv_t0, (list, DataAPIVector))
        assert isinstance(gsv_t0[0], float)

        cur_n1 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
        )
        cur_f1 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_sort_vector=False,
        )
        cur_t1 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_sort_vector=True,
        )
        await cur_n1.__anext__()
        await cur_f1.__anext__()
        await cur_t1.__anext__()

        assert await cur_n1.get_sort_vector() is None
        assert await cur_f1.get_sort_vector() is None
        gsv_t1 = await cur_t1.get_sort_vector()
        assert isinstance(gsv_t1, (list, DataAPIVector))
        assert isinstance(gsv_t1[0], float)

        cur_n2 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
        )
        cur_f2 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_sort_vector=False,
        )
        cur_t2 = acoll.find_and_rerank(
            sort={"$hybrid": {"$vector": [0, 1], "$lexical": "bla"}},
            rerank_on="content",
            rerank_query="blaa",
            include_sort_vector=True,
        )
        await cur_n2.to_list()
        await cur_f2.to_list()
        await cur_t2.to_list()

        assert await cur_n2.get_sort_vector() is None
        assert await cur_f2.get_sort_vector() is None
        gsv_t2 = await cur_t2.get_sort_vector()
        assert isinstance(gsv_t2, (list, DataAPIVector))
        assert isinstance(gsv_t2[0], float)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.api_options import APIOptions, SerdesOptions
from astrapy.constants import DefaultDocumentType
from astrapy.cursors import CursorState, RerankedResult
from astrapy.data_types import DataAPIVector
from astrapy.exceptions import CursorException

from ..conftest import IS_ASTRA_DB, USE_RERANKER_API_KEY_HEADER, DefaultCollection

NUM_DOCS = 25  # keep this between 20 and 39
RUN_RERANKER_TESTS = IS_ASTRA_DB or USE_RERANKER_API_KEY_HEADER


@pytest.fixture
def filled_vectorize_collection(
    sync_empty_farr_vectorize_collection: DefaultCollection,
) -> DefaultCollection:
    sync_empty_farr_vectorize_collection.insert_many(
        [
            {
                "_id": f"doc_{i:02}",
                "$vectorize": "this is sentence #{i} (vecze)",
                "$lexical": "this is sentence #{i} (lexi)",
                "parity": ["even", "odd"][i % 2],
            }
            for i in range(NUM_DOCS)
        ]
    )
    return sync_empty_farr_vectorize_collection


@pytest.mark.skipif(
    not RUN_RERANKER_TESTS,
    reason="Reranker-related tests skipped (no Astra and no header reranker key provided)",
)
class TestCollectionCursorSync:
    @pytest.mark.describe("test of an IDLE collection farr-cursors properties, sync")
    def test_collection_farrcursors_idle_properties_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        assert cur.state == CursorState.IDLE

        assert cur.data_source == filled_vectorize_collection

        assert cur.consumed == 0
        assert cur.consume_buffer(3) == []
        assert cur.buffered_count == 0
        assert cur.consumed == 0

        toclose = cur.clone()
        toclose.close()
        toclose.close()
        assert toclose.state == CursorState.CLOSED
        with pytest.raises(CursorException):
            for row in toclose:
                pass
        with pytest.raises(StopIteration):
            next(toclose)
        with pytest.raises(CursorException):
            toclose.for_each(lambda row: None)
        with pytest.raises(CursorException):
            toclose.to_list()

        cur.rewind()
        assert cur.state == CursorState.IDLE
        assert cur.consumed == 0
        assert cur.buffered_count == 0

        cur.filter({"c": True})
        cur.project({"c": True})
        cur.sort({"$hybrid": "query"})
        cur.limit(1)
        cur.hybrid_limits(2)
        cur.include_sort_vector(False)
        cur.include_scores(True)
        cur.rerank_on("field")
        cur.map(lambda r_res: None)

        cur.project({}).map(lambda r_res: None)
        with pytest.raises(CursorException):
            cur.map(lambda r_res: None).project({})

    @pytest.mark.describe("test of a CLOSED collection farr-cursors properties, sync")
    def test_collection_farrcursors_closed_properties_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        cur0 = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        cur0.close()
        cur0.rewind()
        assert cur0.state == CursorState.IDLE

        cur1 = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        assert cur1.consumed == 0
        list(cur1)
        assert cur1.state == CursorState.CLOSED
        assert cur1.consume_buffer(2) == []
        assert cur1.consumed == NUM_DOCS
        assert cur1.buffered_count == 0
        cloned = cur1.clone()
        assert cloned.consumed == 0
        assert cloned.buffered_count == 0
        assert cloned.state == CursorState.IDLE

        with pytest.raises(CursorException):
            cur1.filter({"c": True})
        with pytest.raises(CursorException):
            cur1.project({"c": True})
        with pytest.raises(CursorException):
            cur1.sort({"$hybrid": "query"})
        with pytest.raises(CursorException):
            cur1.limit(1)
        with pytest.raises(CursorException):
            cur1.hybrid_limits(2)
        with pytest.raises(CursorException):
            cur1.include_scores(False)
        with pytest.raises(CursorException):
            cur1.include_sort_vector(False)
        with pytest.raises(CursorException):
            cur1.rerank_on("field")
        with pytest.raises(CursorException):
            cur1.map(lambda rw: None)

    @pytest.mark.describe("test of a STARTED collection farr-cursors properties, sync")
    def test_collection_farrcursors_started_properties_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        LIMIT = NUM_DOCS - 1
        cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=LIMIT,
        )
        next(cur)
        # now this is a one-page cursor and has LIMIT-1 items in buffer (one gone)
        assert cur.consumed == 1
        assert cur.buffered_count == LIMIT - 1
        assert len(cur.consume_buffer(3)) == 3
        assert cur.consumed == 4
        assert cur.buffered_count == LIMIT - 4
        # from time to time the buffer is empty:
        for _ in range(LIMIT - 4):
            next(cur)
        assert cur.buffered_count == 0
        assert cur.consume_buffer(3) == []
        assert cur.consumed == LIMIT
        assert cur.buffered_count == 0

        with pytest.raises(CursorException):
            cur.filter({"c": True})
        with pytest.raises(CursorException):
            cur.project({"c": True})
        with pytest.raises(CursorException):
            cur.sort({"$hybrid": "query"})
        with pytest.raises(CursorException):
            cur.limit(1)
        with pytest.raises(CursorException):
            cur.hybrid_limits(2)
        with pytest.raises(CursorException):
            cur.include_scores(False)
        with pytest.raises(CursorException):
            cur.include_sort_vector(False)
        with pytest.raises(CursorException):
            cur.rerank_on("field")
        with pytest.raises(CursorException):
            cur.map(lambda rw: None)

    @pytest.mark.describe("test of collection farr-cursors has_next, sync")
    def test_collection_farrcursors_has_next_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        assert cur.state == CursorState.IDLE
        assert cur.consumed == 0
        assert cur.has_next()
        assert cur.state == CursorState.IDLE
        assert cur.consumed == 0
        list(cur)
        assert cur.consumed == NUM_DOCS
        assert cur.state == CursorState.CLOSED  # type: ignore[comparison-overlap]

        curmf = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        next(curmf)
        next(curmf)
        assert curmf.consumed == 2
        assert curmf.state == CursorState.STARTED
        assert curmf.has_next()
        assert curmf.consumed == 2
        assert curmf.state == CursorState.STARTED
        for _ in range(18):
            next(curmf)
        # not very relevant as there's no actual pages here
        assert curmf.has_next()
        assert curmf.consumed == 20
        assert curmf.state == CursorState.STARTED
        assert curmf.buffered_count == NUM_DOCS - 20

        cur0 = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        cur0.close()
        assert not cur0.has_next()

    @pytest.mark.describe("test of collection farr-cursors zero matches, sync")
    def test_collection_farrcursors_zeromatches_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        cur = filled_vectorize_collection.find_and_rerank(
            {"parity": -1},
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        assert not cur.has_next()
        assert list(cur) == []

        cur_no_sv = filled_vectorize_collection.find_and_rerank(
            {"parity": -1},
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
            include_sort_vector=False,
        )
        assert cur_no_sv.get_sort_vector() is None
        assert list(cur_no_sv) == []
        assert cur_no_sv.get_sort_vector() is None

        cur_with_sv = filled_vectorize_collection.find_and_rerank(
            {"parity": -1},
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
            include_sort_vector=True,
        )
        sort_vector = cur_with_sv.get_sort_vector()
        assert sort_vector is not None
        assert isinstance(sort_vector, DataAPIVector)

    @pytest.mark.describe("test of prematurely farr-closing collection cursors, sync")
    def test_collection_farrcursors_early_closing_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        for _ in range(12):
            next(cur)
        cur.close()
        assert cur.state == CursorState.CLOSED
        assert cur.buffered_count == 0
        assert cur.consumed == 12
        # rewind test
        cur.rewind()
        assert len(list(cur)) == NUM_DOCS

    @pytest.mark.describe("test of mappings with collection farr-cursors, sync")
    def test_collection_farrcursors_mapping_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        base_rows = list(
            filled_vectorize_collection.find_and_rerank(
                sort={"$hybrid": "a sentence."},
                limit=NUM_DOCS,
            )
        )
        assert len(base_rows) == NUM_DOCS
        base_rows_mu = list(
            filled_vectorize_collection.find_and_rerank(
                sort={"$hybrid": "a sentence."},
                limit=NUM_DOCS - 3,
            )
        )
        assert len(base_rows_mu) == NUM_DOCS - 3

        parity_backmap = {"odd": 1, "even": 0}

        def mint(r_result: RerankedResult[DefaultDocumentType]) -> int:
            return 1 + parity_backmap[r_result.document["parity"]]

        def mmult(val: int) -> int:
            return 1000 * val

        # map, base
        mcur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        ).map(mint)
        mints = [val for val in mcur]
        assert mints == [mint(row) for row in base_rows]

        # map composition
        mmcur = (
            filled_vectorize_collection.find_and_rerank(
                sort={"$hybrid": "a sentence."},
                limit=NUM_DOCS,
            )
            .map(mint)
            .map(mmult)
        )
        mmints = [val for val in mmcur]
        assert mmints == [mmult(mint(row)) for row in base_rows]

        # consume_buffer skips the map
        hmcur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        ).map(mint)
        for _ in range(10):
            next(hmcur)
        conbuf = hmcur.consume_buffer(2)
        assert len(conbuf) == 2
        assert all(isinstance(itm, RerankedResult) for itm in conbuf)

        # rewind preserves the mapping
        rwcur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        ).map(mint)
        for _ in range(10):
            next(rwcur)
        rwcur.rewind()
        assert next(rwcur) == mints[0]

        # clone rewinds
        cl_unmapped = rwcur.clone()
        assert next(cl_unmapped) == mint(base_rows[0])

    @pytest.mark.describe("test of collection farr-cursors, for_each and to_list, sync")
    def test_collection_farrcursors_collective_methods_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        base_rows = list(
            filled_vectorize_collection.find_and_rerank(
                sort={"$hybrid": "a sentence."},
                limit=NUM_DOCS,
            )
        )

        # full to_list
        tl_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        assert tl_cur.to_list() == base_rows
        assert tl_cur.state == CursorState.CLOSED

        # partially-consumed to_list
        ptl_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        for _ in range(15):
            next(ptl_cur)
        assert ptl_cur.to_list() == base_rows[15:]
        assert ptl_cur.state == CursorState.CLOSED

        # mapped to_list

        parity_backmap = {"odd": 1, "even": 0}

        def mint(r_result: RerankedResult[DefaultDocumentType]) -> int:
            return parity_backmap[r_result.document["parity"]]

        mtl_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        ).map(mint)
        for _ in range(13):
            next(mtl_cur)
        assert mtl_cur.to_list() == [mint(row) for row in base_rows[13:]]
        assert mtl_cur.state == CursorState.CLOSED

        # full for_each
        accum0: list[RerankedResult[DefaultDocumentType]] = []

        def marker0(
            r_result: RerankedResult[DefaultDocumentType],
            acc: list[RerankedResult[DefaultDocumentType]] = accum0,
        ) -> None:
            acc += [r_result]

        fe_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        fe_cur.for_each(marker0)
        assert accum0 == base_rows
        assert fe_cur.state == CursorState.CLOSED

        # partially-consumed for_each
        accum1: list[RerankedResult[DefaultDocumentType]] = []

        def marker1(
            r_result: RerankedResult[DefaultDocumentType],
            acc: list[RerankedResult[DefaultDocumentType]] = accum1,
        ) -> None:
            acc += [r_result]

        pfe_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        for _ in range(11):
            next(pfe_cur)
        pfe_cur.for_each(marker1)
        assert accum1 == base_rows[11:]
        assert pfe_cur.state == CursorState.CLOSED

        # mapped for_each
        accum2: list[int] = []

        def marker2(val: int, acc: list[int] = accum2) -> None:
            acc += [val]

        mfe_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        ).map(mint)
        for _ in range(17):
            next(mfe_cur)
        mfe_cur.for_each(marker2)
        assert accum2 == [mint(row) for row in base_rows[17:]]
        assert mfe_cur.state == CursorState.CLOSED

        # breaking (early) for_each
        accum3: list[RerankedResult[DefaultDocumentType]] = []

        def marker3(
            r_result: RerankedResult[DefaultDocumentType],
            acc: list[RerankedResult[DefaultDocumentType]] = accum3,
        ) -> bool:
            acc += [r_result]
            return len(acc) < 5

        bfe_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        bfe_cur.for_each(marker3)
        assert accum3 == base_rows[:5]
        assert bfe_cur.state == CursorState.STARTED
        bfe_another = next(bfe_cur)
        assert bfe_another == base_rows[5]

        # nonbool-nonbreaking for_each
        accum4: list[RerankedResult[DefaultDocumentType]] = []

        def marker4(
            r_result: RerankedResult[DefaultDocumentType],
            acc: list[RerankedResult[DefaultDocumentType]] = accum4,
        ) -> int:
            acc += [r_result]
            return 8 if len(acc) < 5 else 0

        nbfe_cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=NUM_DOCS,
        )
        nbfe_cur.for_each(marker4)  # type: ignore[arg-type]
        assert accum4 == base_rows
        assert nbfe_cur.state == CursorState.CLOSED

    @pytest.mark.describe(
        "test of collection farr-cursors, serdes options obeyance, sync"
    )
    def test_collection_farrcursors_serdes_options_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        # serdes options obeyance check
        noncustom_compo_collection = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            )
        )
        custom_compo_collection = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            )
        )
        noncustom_rows = noncustom_compo_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            projection={"$vector": True},
            limit=NUM_DOCS,
        ).to_list()
        assert len(noncustom_rows) == NUM_DOCS
        assert all(
            isinstance(nc_r_res.document["$vector"], list)
            for nc_r_res in noncustom_rows
        )
        custom_rows = custom_compo_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            projection={"$vector": True},
            limit=NUM_DOCS,
        ).to_list()
        assert len(custom_rows) == NUM_DOCS
        assert all(
            isinstance(c_r_res.document["$vector"], DataAPIVector)
            for c_r_res in custom_rows
        )

    @pytest.mark.describe(
        "test of collection farr-cursors, include_sort_vector and serdes, sync"
    )
    def test_collection_farrcursors_include_sort_vector_serdes_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        col_v0_d0 = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=False,
                    use_decimals_in_collections=False,
                ),
            ),
        )
        col_v0_d1 = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=False,
                    use_decimals_in_collections=True,
                ),
            ),
        )
        col_v1_d0 = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=True,
                    use_decimals_in_collections=False,
                ),
            ),
        )
        col_v1_d1 = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    custom_datatypes_in_reading=True,
                    use_decimals_in_collections=True,
                ),
            ),
        )

        cur0_v0_d0 = col_v0_d0.find_and_rerank(sort={"$hybrid": "query"})
        cur0_v0_d1 = col_v0_d1.find_and_rerank(sort={"$hybrid": "query"})
        cur0_v1_d0 = col_v1_d0.find_and_rerank(sort={"$hybrid": "query"})
        cur0_v1_d1 = col_v1_d1.find_and_rerank(sort={"$hybrid": "query"})
        assert cur0_v0_d0.get_sort_vector() is None
        assert cur0_v0_d1.get_sort_vector() is None
        assert cur0_v1_d0.get_sort_vector() is None
        assert cur0_v1_d1.get_sort_vector() is None

        cur1_v0_d0 = col_v0_d0.find_and_rerank(
            sort={"$hybrid": "query"},
            include_sort_vector=True,
        )
        cur1_v0_d1 = col_v0_d1.find_and_rerank(
            sort={"$hybrid": "query"},
            include_sort_vector=True,
        )
        cur1_v1_d0 = col_v1_d0.find_and_rerank(
            sort={"$hybrid": "query"},
            include_sort_vector=True,
        )
        cur1_v1_d1 = col_v1_d1.find_and_rerank(
            sort={"$hybrid": "query"},
            include_sort_vector=True,
        )
        sv_v0_d0 = cur1_v0_d0.get_sort_vector()
        sv_v0_d1 = cur1_v0_d1.get_sort_vector()
        sv_v1_d0 = cur1_v1_d0.get_sort_vector()
        sv_v1_d1 = cur1_v1_d1.get_sort_vector()

        assert isinstance(sv_v0_d0, list)
        assert isinstance(sv_v0_d1, list)
        assert isinstance(sv_v0_d0[0], float)
        assert isinstance(sv_v0_d1[0], float)
        assert isinstance(sv_v1_d0, DataAPIVector)
        assert isinstance(sv_v1_d1, DataAPIVector)

    @pytest.mark.describe("test of farr collection cursors fetch_next_page, sync")
    def test_collection_farrcursors_fetchnextpage_sync(
        self,
        filled_vectorize_collection: DefaultCollection,
    ) -> None:
        # A very limited test since FARR is one-page-only
        cur = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=5,
        )
        cur_fp = filled_vectorize_collection.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=5,
        )
        lst = cur.to_list()
        page_fp = cur_fp.fetch_next_page()
        assert page_fp.next_page_state is None
        assert page_fp.sort_vector is None
        docs_from_cursor = [rr.document for rr in lst]
        docs_from_page = [rr_fp.document for rr_fp in page_fp.results]
        assert docs_from_cursor == docs_from_page

        # rewind behaviour
        cur.rewind(initial_page_state="some string")
        cur.rewind()
        with pytest.raises(ValueError, match="null"):
            cur.rewind(initial_page_state=None)  # type: ignore[arg-type]

        # forbidden: mixing pagination and ordinary usage
        cur.__next__()
        with pytest.raises(CursorException):
            cur.fetch_next_page()
        cur.to_list()
        with pytest.raises(CursorException):
            cur.fetch_next_page()

        # behaviour re: include_sort_vector and its format
        coll_noncustom = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            )
        )
        coll_custom = filled_vectorize_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            )
        )

        vcur0_nc = coll_noncustom.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=5,
            include_sort_vector=True,
        )
        vpage0_nc = vcur0_nc.fetch_next_page()
        assert vpage0_nc.next_page_state is None
        assert len(vpage0_nc.results) == 5
        assert isinstance(vpage0_nc.sort_vector, list)
        assert not isinstance(vpage0_nc.sort_vector, DataAPIVector)

        vcur0_c = coll_custom.find_and_rerank(
            sort={"$hybrid": "a sentence."},
            limit=5,
            include_sort_vector=True,
        )
        vpage0_c = vcur0_c.fetch_next_page()
        assert vpage0_c.next_page_state is None
        assert len(vpage0_c.results) == 5
        assert isinstance(vpage0_c.sort_vector, DataAPIVector)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.data_types import DataAPIMap, DataAPISet

from ..conftest import IS_ASTRA_DB, DefaultTable


class TestTableCollectionColumnsSync:
    @pytest.mark.describe("test of table list and set columns, filtering, sync")
    def test_table_listset_columns_filtering_sync(
        self,
        sync_empty_table_collindexed: DefaultTable,
    ) -> None:
        table = sync_empty_table_collindexed
        table.insert_one(
            {
                "id": "bls_low",
                "set_int": DataAPISet([10, 11, 12]),
                "list_int": [20, 21, 22],
            }
        )
        table.insert_many(
            [
                {
                    "id": "bls_high",
                    "set_int": [110, 111, 112],
                    "list_int": [120, 121, 122],
                }
            ]
        )

        res_s0 = table.find_one({"set_int": {"$in": [11]}})
        assert res_s0 is not None
        assert res_s0["id"] == "bls_low"
        assert table.find({"set_int": {"$in": [11]}}, limit=1).to_list()[0] == res_s0

        res_s1 = table.find_one({"set_int": {"$in": [11, 99]}})
        assert res_s1 is not None
        assert res_s1["id"] == "bls_low"
        assert (
            table.find({"set_int": {"$in": [11, 99]}}, limit=1).to_list()[0] == res_s1
        )

        res_s0n = table.find_one({"set_int": {"$nin": [111]}})
        assert res_s0n is not None
        assert res_s0n["id"] == "bls_low"
        assert table.find({"set_int": {"$nin": [111]}}, limit=1).to_list()[0] == res_s0n

        res_s1n = table.find_one({"set_int": {"$nin": [10, 110]}})
        assert res_s1n is None
        assert table.find({"set_int": {"$nin": [10, 110]}}, limit=1).to_list() == []

        res_s2 = table.find_one({"set_int": {"$all": [11, 12]}})
        assert res_s2 is not None
        assert res_s2["id"] == "bls_low"
        assert (
            table.find({"set_int": {"$all": [11, 12]}}, limit=1).to_list()[0] == res_s2
        )

        res_s3 = table.find_one({"set_int": {"$all": [11, 99]}})
        assert res_s3 is None
        assert table.find({"set_int": {"$all": [11, 99]}}, limit=1).to_list() == []

        res_l0 = table.find_one({"list_int": {"$in": [21]}})
        assert res_l0 is not None
        assert res_l0["id"] == "bls_low"
        assert table.find({"list_int": {"$in": [21]}}, limit=1).to_list()[0] == res_l0

        res_l1 = table.find_one({"list_int": {"$in": [21, 99]}})
        assert res_l1 is not None
        assert res_l1["id"] == "bls_low"
        assert (
            table.find({"list_int": {"$in": [21, 99]}}, limit=1).to_list()[0] == res_l1
        )

        res_l0n = table.find_one({"list_int": {"$nin": [120]}})
        assert res_l0n is not None
        assert res_l0n["id"] == "bls_low"
        assert (
            table.find({"list_int": {"$nin": [120]}}, limit=1).to_list()[0] == res_l0n
        )

        res_l1n = table.find_one({"list_int": {"$nin": [20, 120]}})
        assert res_l1n is None
        assert table.find({"list_int": {"$nin": [20, 120]}}, limit=1).to_list() == []

        res_l2 = table.find_one({"list_int": {"$all": [21, 22]}})
        assert res_l2 is not None
        assert res_l2["id"] == "bls_low"
        assert (
            table.find({"list_int": {"$all": [21, 22]}}, limit=1).to_list()[0] == res_l2
        )

        res_l3 = table.find_one({"list_int": {"$all": [21, 99]}})
        assert res_l3 is None
        assert table.find({"list_int": {"$all": [21, 99]}}, limit=1).to_list() == []

    @pytest.mark.describe("test of table map columns, filtering, sync")
    def test_table_map_columns_filtering_sync(
        self,
        sync_empty_table_collindexed: DefaultTable,
    ) -> None:
        table = sync_empty_table_collindexed
        table.insert_one(
            {
                "id": "bm_low",
                "map_text_int_e": DataAPIMap({"a": 1, "b": 2, "c": 3}),
                "map_text_int_k": DataAPIMap({"A": 1, "B": 2, "C": 3}),
                "map_text_int_v": DataAPIMap({"x": 11, "y": 12, "z": 13}),
            }
        )
        table.insert_many(
            [
                {
                    "id": "bm_high",
                    "map_text_int_e": DataAPIMap({"a": 101, "b": 102, "c": 103}),
                    "map_text_int_k": DataAPIMap({"A": 101, "B": 102, "C2": 103}),
                    "map_text_int_v": DataAPIMap({"x": 101, "y": 102, "z": 103}),
                }
            ]
        )

        res_me0 = table.find_one({"map_text_int_e": {"$in": [["a", 1]]}})
        assert res_me0 is not None
        assert res_me0["id"] == "bm_low"
        assert (
            table.find({"map_text_int_e": {"$in": [["a", 1]]}}, limit=1).to_list()[0]
            == res_me0
        )

        res_me1 = table.find_one({"map_text_int_e": {"$in": [["a", 1], ["q", 9]]}})
        assert res_me1 is not None
        assert res_me1["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_e": {"$in": [["a", 1], ["q", 9]]}}, limit=1
            ).to_list()[0]
            == res_me1
        )

        res_me0n = table.find_one({"map_text_int_e": {"$nin": [["a", 101]]}})
        assert res_me0n is not None
        assert res_me0n["id"] == "bm_low"
        assert (
            table.find({"map_text_int_e": {"$nin": [["a", 101]]}}, limit=1).to_list()[0]
            == res_me0n
        )

        res_me1n = table.find_one({"map_text_int_e": {"$nin": [["a", 101], ["a", 1]]}})
        assert res_me1n is None
        assert (
            table.find(
                {"map_text_int_e": {"$nin": [["a", 101], ["a", 1]]}}, limit=1
            ).to_list()
            == []
        )

        res_me2 = table.find_one({"map_text_int_e": {"$all": [["a", 1], ["b", 2]]}})
        assert res_me2 is not None
        assert res_me2["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_e": {"$all": [["a", 1], ["b", 2]]}}, limit=1
            ).to_list()[0]
            == res_me2
        )

        res_me3 = table.find_one({"map_text_int_e": {"$all": [["a", 1], ["q", 9]]}})
        assert res_me3 is None
        assert (
            table.find(
                {"map_text_int_e": {"$all": [["a", 1], ["q", 9]]}}, limit=1
            ).to_list()
            == []
        )

        res_mk0 = table.find_one({"map_text_int_k": {"$keys": {"$in": ["A"]}}})
        assert res_mk0 is not None
        assert res_mk0["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_k": {"$keys": {"$in": ["A"]}}}, limit=1
            ).to_list()[0]
            == res_mk0
        )

        res_mk1 = table.find_one({"map_text_int_k": {"$keys": {"$in": ["A", "Q"]}}})
        assert res_mk1 is not None
        assert res_mk1["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_k": {"$keys": {"$in": ["A", "Q"]}}}, limit=1
            ).to_list()[0]
            == res_mk1
        )

        res_mk0n = table.find_one({"map_text_int_k": {"$keys": {"$nin": ["C2"]}}})
        assert res_mk0n is not None
        assert res_mk0n["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_k": {"$keys": {"$nin": ["C2"]}}}, limit=1
            ).to_list()[0]
            == res_mk0n
        )

        res_mk1n = table.find_one({"map_text_int_k": {"$keys": {"$nin": ["C", "C2"]}}})
        assert res_mk1n is None
        assert (
            table.find(
                {"map_text_int_k": {"$keys": {"$nin": ["A", "Q"]}}}, limit=1
            ).to_list()
            == []
        )

        res_mk2 = table.find_one({"map_text_int_k": {"$keys": {"$all": ["A", "B"]}}})
        assert res_mk2 is not None
        assert res_mk2["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_k": {"$keys": {"$all": ["A", "B"]}}}, limit=1
            ).to_list()[0]
            == res_mk2
        )

        res_mk3 = table.find_one({"map_text_int_k": {"$keys": {"$all": ["A", "Q"]}}})
        assert res_mk3 is None
        assert (
            table.find(
                {"map_text_int_k": {"$keys": {"$all": ["A", "Q"]}}}, limit=1
            ).to_list()
            == []
        )

        res_mv0 = table.find_one({"map_text_int_v": {"$values": {"$in": [11]}}})
        assert res_mv0 is not None
        assert res_mv0["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_v": {"$values": {"$in": [11]}}}, limit=1
            ).to_list()[0]
            == res_mv0
        )

        res_mv1 = table.find_one({"map_text_int_v": {"$values": {"$in": [11, 99]}}})
        assert res_mv1 is not None
        assert res_mv1["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_v": {"$values": {"$in": [11, 99]}}}, limit=1
            ).to_list()[0]
            == res_mv1
        )

        res_mv0n = table.find_one({"map_text_int_v": {"$values": {"$nin": [101]}}})
        assert res_mv0n is not None
        assert res_mv0n["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_v": {"$values": {"$nin": [101]}}}, limit=1
            ).to_list()[0]
            == res_mv0n
        )

        res_mv1n = table.find_one({"map_text_int_v": {"$values": {"$nin": [101, 11]}}})
        assert res_mv1n is None
        assert (
            table.find(
                {"map_text_int_v": {"$values": {"$nin": [101, 11]}}}, limit=1
            ).to_list()
            == []
        )

        res_mv2 = table.find_one({"map_text_int_v": {"$values": {"$all": [11, 12]}}})
        assert res_mv2 is not None
        assert res_mv2["id"] == "bm_low"
        assert (
            table.find(
                {"map_text_int_v": {"$values": {"$all": [11, 12]}}}, limit=1
            ).to_list()[0]
            == res_mv2
        )

        res_mv3 = table.find_one({"map_text_int_v": {"$values": {"$all": [11, 99]}}})
        assert res_mv3 is None
        assert (
            table.find(
                {"map_text_int_v": {"$values": {"$all": [11, 99]}}}, limit=1
            ).to_list()
            == []
        )

    @pytest.mark.skipif(IS_ASTRA_DB, reason="Astra rejects reads-before-writes.")
    @pytest.mark.describe("test of table list and set columns, dollar-writes, sync")
    def test_table_listset_columns_dollarwriting_sync(
        self,
        sync_empty_table_collindexed: DefaultTable,
    ) -> None:
        table = sync_empty_table_collindexed
        table.insert_one(
            {
                "id": "dwr_bls",
                "set_int": DataAPISet([30, 31, 32]),
                "list_int": [40, 41, 42],
            }
        )

        table.update_one(
            {"id": "dwr_bls"},
            {"$pullAll": {"set_int": [32, 30]}},
        )
        post_pa_s = table.find_one({"id": "dwr_bls"})
        assert post_pa_s is not None
        assert post_pa_s["set_int"] == DataAPISet([31])

        table.update_one(
            {"id": "dwr_bls"},
            {"$pullAll": {"list_int": [42, 40]}},
        )
        post_pa_l = table.find_one({"id": "dwr_bls"})
        assert post_pa_l is not None
        assert post_pa_l["list_int"] == [41]

        table.update_one(
            {"id": "dwr_bls"},
            {
                "$pullAll": {
                    "set_int": [31],
                    "list_int": [41],
                },
            },
        )
        post_pa_ls = table.find_one({"id": "dwr_bls"})
        assert post_pa_ls is not None
        assert post_pa_ls["list_int"] == []
        assert post_pa_ls["set_int"] == DataAPISet()

        table.update_one(
            {"id": "dwr_bls"},
            {
                "$push": {
                    "set_int": 930,
                }
            },
        )
        post_pu_s = table.find_one({"id": "dwr_bls"})
        assert post_pu_s is not None
        assert post_pu_s["list_int"] == []
        assert post_pu_s["set_int"] == DataAPISet([930])

        table.update_one(
            {"id": "dwr_bls"},
            {
                "$push": {
                    "list_int": 940,
                }
            },
        )
        post_pu_l = table.find_one({"id": "dwr_bls"})
        assert post_pu_l is not None
        assert post_pu_l["list_int"] == [940]
        assert post_pu_l["set_int"] == DataAPISet([930])

        table.update_one(
            {"id": "dwr_bls"},
            {
                "$push": {
                    "set_int": {
                        "$each": [931, 932],
                    },
                    "list_int": {"$each": [941, 942]},
                }
            },
        )
        post_pu_ls = table.find_one({"id": "dwr_bls"})
        assert post_pu_ls is not None
        assert post_pu_ls["list_int"] == [940, 941, 942]
        assert post_pu_ls["set_int"] == DataAPISet([930, 931, 932])

    @pytest.mark.describe("test of table map columns, dollar-writes, sync")
    def test_table_map_columns_dollarwriting_sync(
        self,
        sync_empty_table_collindexed: DefaultTable,
    ) -> None:
        table = sync_empty_table_collindexed
        table.insert_one(
            {
                "id": "dwr_m",
                "map_text_int_e": DataAPIMap({"a": 1, "b": 2, "c": 3}),
            }
        )

        table.update_one(
            {"id": "dwr_m"},
            {
                "$pullAll": {
                    "map_text_int_e": ["b", "c"],
                },
            },
        )
        post_pa_m = table.find_one({"id": "dwr_m"})
        assert post_pa_m is not None
        assert post_pa_m["map_text_int_e"] == DataAPIMap({"a": 1})

        table.update_one(
            {"id": "dwr_m"},
            {
                "$pullAll": {
                    "map_text_int_e": ["a", "b"],
                },
            },
        )
        post_pa_m2 = table.find_one({"id": "dwr_m"})
        assert post_pa_m2 is not None
        assert post_pa_m2["map_text_int_e"] == DataAPIMap({})

        table.update_one(
            {"id": "dwr_m"},
            {
                "$push": {
                    "map_text_int_e": ["z", 60],
                },
            },
        )
        post_pu_m = table.find_one({"id": "dwr_m"})
        assert post_pu_m is not None
        assert post_pu_m["map_text_int_e"] == DataAPIMap({"z": 60})

        table.update_one(
            {"id": "dwr_m"},
            {
                "$push": {
                    "map_text_int_e": {"y": 61},
                },
            },
        )
        post_pu_m2 = table.find_one({"id": "dwr_m"})
        assert post_pu_m2 is not None
        assert post_pu_m2["map_text_int_e"] == DataAPIMap({"z": 60, "y": 61})

        table.update_one(
            {"id": "dwr_m"},
            {
                "$push": {
                    "map_text_int_e": {
                        "$each": [
                            {"x": 62},
                            {"w": 63},
                        ]
                    },
                },
            },
        )
        post_pu_m3 = table.find_one({"id": "dwr_m"})
        assert post_pu_m3 is not None
        assert post_pu_m3["map_text_int_e"] == DataAPIMap(
            {"z": 60, "y": 61, "x": 62, "w": 63},
        )

        table.update_one(
            {"id": "dwr_m"},
            {
                "$push": {
                    "map_text_int_e": {
                        "$each": [
                            ["v", 64],
                            ["u", 65],
                        ]
                    },
                },
            },
        )
        post_pu_m4 = table.find_one({"id": "dwr_m"})
        assert post_pu_m4 is not None
        assert post_pu_m4["map_text_int_e"] == DataAPIMap(
            {"z": 60, "y": 61, "x": 62, "w": 63, "v": 64, "u": 65},
        )

        table.update_one(
            {"id": "dwr_m"},
            {
                "$set": {
                    "map_text_int_k": {
                        "r": 70,
                        "s": 71,
                    },
                },
            },
        )
        post_se_m = table.find_one({"id": "dwr_m"})
        assert post_se_m is not None
        assert post_se_m["map_text_int_k"] == DataAPIMap({"r": 70, "s": 71})

        table.update_one(
            {"id": "dwr_m"},
            {
                "$set": {
                    "map_int_text": [
                        [72, "r2"],
                        [73, "s2"],
                    ],
                },
            },
        )
        post_se_m2 = table.find_one({"id": "dwr_m"})
        assert post_se_m2 is not None
        assert post_se_m2["map_int_text"] == DataAPIMap({72: "r2", 73: "s2"})

        table.update_one(
            {"id": "dwr_m"},
            {
                "$set": {
                    "map_int_text": DataAPIMap(
                        [
                            (74, "r3"),
                            (75, "s3"),
                        ]
                    ),
                },
            },
        )
        post_se_m3 = table.find_one({"id": "dwr_m"})
        assert post_se_m3 is not None
        assert post_se_m3["map_int_text"] == DataAPIMap({74: "r3", 75: "s3"})



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime

# TODO NOINETINUDT
# import ipaddress
from dataclasses import dataclass
from typing import Any

from astrapy.data.utils.extended_json_converters import convert_to_ejson_bytes
from astrapy.data_types import DataAPITimestamp
from astrapy.info import (
    CreateTableDefinition,
    CreateTypeDefinition,
    TableKeyValuedColumnTypeDescriptor,
    TablePrimaryKeyDescriptor,
    TableScalarColumnTypeDescriptor,
    TableUDTColumnDescriptor,
    TableValuedColumnTypeDescriptor,
)

PLAYER_TYPE_NAME = "udt_player"
PLAYER_TYPE_DEFINITION = CreateTypeDefinition(
    fields={
        "name": "text",
        "age": "int",
    },
)
PLAYER_TABLE_NAME = "test_table_udt_player"
PLAYER_TABLE_DEFINITION = CreateTableDefinition.coerce(
    {
        "columns": {
            "id": "text",
            "scalar_udt": {"type": "userDefined", "udtName": PLAYER_TYPE_NAME},
            "list_udt": {
                "type": "list",
                "valueType": {
                    "type": "userDefined",
                    "udtName": PLAYER_TYPE_NAME,
                },
            },
            "set_udt": {
                "type": "set",
                "valueType": {
                    "type": "userDefined",
                    "udtName": PLAYER_TYPE_NAME,
                },
            },
            "map_udt": {
                "type": "map",
                "keyType": "text",
                "valueType": {
                    "type": "userDefined",
                    "udtName": PLAYER_TYPE_NAME,
                },
            },
        },
        "primaryKey": {
            "partitionBy": ["id"],
            "partitionSort": {},
        },
    }
)

EXTENDED_PLAYER_TYPE_NAME = "udt_extended_player"
EXTENDED_PLAYER_TYPE_DEFINITION = CreateTypeDefinition(
    fields={
        "name": "text",
        "age": "int",
        "ts": "timestamp",
        # TODO NOINETINUDT
        # "inet": "inet",
    },
)
#        TODO NOBLOBINUDT "blb": "blob",
EXTENDED_PLAYER_TABLE_NAME = "test_table_udt_extended_player"
EXTENDED_PLAYER_TABLE_DEFINITION = CreateTableDefinition.coerce(
    {
        "columns": {
            "id": "text",
            "scalar_udt": {"type": "userDefined", "udtName": EXTENDED_PLAYER_TYPE_NAME},
            "list_udt": {
                "type": "list",
                "valueType": {
                    "type": "userDefined",
                    "udtName": EXTENDED_PLAYER_TYPE_NAME,
                },
            },
            "set_udt": {
                "type": "set",
                "valueType": {
                    "type": "userDefined",
                    "udtName": EXTENDED_PLAYER_TYPE_NAME,
                },
            },
            "map_udt": {
                "type": "map",
                "keyType": "text",
                "valueType": {
                    "type": "userDefined",
                    "udtName": EXTENDED_PLAYER_TYPE_NAME,
                },
            },
        },
        "primaryKey": {
            "partitionBy": ["id"],
            "partitionSort": {},
        },
    }
)

UNIT_EXTENDED_PLAYER_TYPE_NAME = "unit_udt_extended_player"
UNIT_EXTENDED_PLAYER_TYPE_DEFINITION = CreateTypeDefinition(
    fields={
        "name": "text",
        "age": "int",
        "blb": "blob",
        "ts": "timestamp",
    },
)

UNIT_OPTLST_PLAYER_TYPE_NAME = "unit_udt_nullable_requiring_player"
UNIT_OPTLST_PLAYER_TYPE_DEFINITION = CreateTypeDefinition(
    fields={
        "name": "text",
        "age": "int",
        "victories": {
            "type": "list",
            "valueType": "text",
        },
    },
)

TEST_UDTCOLLINDEXED_TABLE_NAME = "test_table_udtcollindexed"
TEST_UDTCOLLINDEXED_TABLE_DEFINITION = CreateTableDefinition(
    columns={
        "id": TableScalarColumnTypeDescriptor(column_type="text"),
        "scalar_udt": TableUDTColumnDescriptor(udt_name=PLAYER_TYPE_NAME),
        "set_udt": TableValuedColumnTypeDescriptor(
            column_type="set",
            value_type=TableUDTColumnDescriptor(udt_name=PLAYER_TYPE_NAME),
        ),
        "list_udt": TableValuedColumnTypeDescriptor(
            column_type="list",
            value_type=TableUDTColumnDescriptor(udt_name=PLAYER_TYPE_NAME),
        ),
        "map_text_udt_e": TableKeyValuedColumnTypeDescriptor(
            key_type="text",
            value_type=TableUDTColumnDescriptor(udt_name=PLAYER_TYPE_NAME),
        ),
        "map_text_udt_v": TableKeyValuedColumnTypeDescriptor(
            key_type="text",
            value_type=TableUDTColumnDescriptor(udt_name=PLAYER_TYPE_NAME),
        ),
    },
    primary_key=TablePrimaryKeyDescriptor(
        partition_by=["id"],
        partition_sort={},
    ),
)
TEST_UDTCOLLINDEXED_TABLE_INDEXES: list[tuple[str, str | dict[str, str]]] = [
    ("test_udtcollidxtable_idx_sudt", "set_udt"),
    ("test_udtcollidxtable_idx_ludt", "list_udt"),
    ("test_udtcollidxtable_idx_mtudte", "map_text_udt_e"),
    ("test_udtcollidxtable_idx_mtudtv", {"map_text_udt_v": "$values"}),
]

THE_BYTES = b"\xa6"
THE_SERIALIZED_BYTES = convert_to_ejson_bytes(THE_BYTES)
THE_TIMESTAMP = DataAPITimestamp.from_string("2025-10-29T01:25:37.123Z")
THE_SERIALIZED_TIMESTAMP = THE_TIMESTAMP.to_string()
THE_TIMEZONE = datetime.timezone(datetime.timedelta(hours=2, minutes=45))
THE_DATETIME = THE_TIMESTAMP.to_datetime(tz=THE_TIMEZONE)


@dataclass
class Player:
    """
    An example dataclass which may be used to represent a user-defined type (UDT)
    such as one would define, and create on the database, with this code:

    .. code-block:: python

        from astrapy.info import CreateTypeDefinition, ColumnType

        player_udt_def = CreateTypeDefinition(fields={
            "name": ColumnType.TEXT,
            "age": ColumnType.INT,
        })

        database.create_type("player_udt", definition=player_udt_def)
    """

    name: str
    age: int


@dataclass
class ExtendedPlayer:
    """
    An example dataclass which may be used to represent a user-defined type (UDT)
    such as one would define, and create on the database, with this code:

    .. code-block:: python

        from astrapy.info import CreateTypeDefinition, ColumnType

        xplayer_udt_def = CreateTypeDefinition(fields={
            "name": ColumnType.TEXT,
            "age": ColumnType.INT,
            "inet": ColumnType.INET,
            "blb": ColumnType.BLOB,
            "ts": ColumnType.TIMESTAMP,
        })

        database.create_type("xplayer_udt", definition=xplayer_udt_def)

    See the test functions using this resource for actual usage examples.
    """

    name: str
    age: int
    # TODO NOINETINUDT
    # inet: ipaddress.IPv4Address | ipaddress.IPv6Address
    # TODO NOBLOBINUDT blb: bytes
    ts: DataAPITimestamp | datetime.datetime


# TODO NOBLOBINUDT reabsorb this (+usage in Unit Tests) once reinstating blb above
@dataclass
class UnitExtendedPlayer:
    """
    An example dataclass which may be used to represent a user-defined type (UDT)
    such as one would define, and create on the database, with this code:

    .. code-block:: python

        from astrapy.info import CreateTypeDefinition, ColumnType

        xplayer_udt_def = CreateTypeDefinition(fields={
            "name": ColumnType.TEXT,
            "age": ColumnType.INT,
            "blb": ColumnType.BLOB,
            "ts": ColumnType.TIMESTAMP,
        })

        database.create_type("xplayer_udt", definition=xplayer_udt_def)

    See the test functions using this resource for actual usage examples.
    """

    name: str
    age: int
    blb: bytes
    ts: DataAPITimestamp | datetime.datetime


@dataclass
class NullablePlayer:
    """
    A counterpart of the Player model class (see), but whose each field
    can be omitted and defaults to None.
    """

    name: str | None = None
    age: int | None = None


@dataclass
class UnitNullableRequiringPlayer:
    """
    A counterpart of the Player model class (see), but:
    1. whose each field admit None, while still being mandatory;
    2. with a field whose "filler" is not null, rather something nontrivial.
    """

    name: str | None
    age: int | None
    victories: list[str]


def _extended_player_serializer(uexp: ExtendedPlayer) -> dict[str, Any]:
    return {k: v for k, v in uexp.__dict__.items() if v is not None}


def _nullable_player_serializer(np: NullablePlayer) -> dict[str, Any]:
    return {k: v for k, v in np.__dict__.items() if v is not None}


def _player_serializer(np: Player) -> dict[str, Any]:
    return {k: v for k, v in np.__dict__.items() if v is not None}


def _unit_extended_player_serializer(uexp: UnitExtendedPlayer) -> dict[str, Any]:
    return {k: v for k, v in uexp.__dict__.items() if v is not None}


def _extended_player_from_dict(
    udict: dict[str, Any],
    definition: CreateTypeDefinition | None,
) -> ExtendedPlayer:
    return ExtendedPlayer(**udict)


def _nullable_player_from_dict(
    udict: dict[str, Any],
    definition: CreateTypeDefinition | None,
) -> NullablePlayer:
    return NullablePlayer(**udict)


def _unit_optlst_player_from_dict(
    udict: dict[str, Any],
    definition: CreateTypeDefinition | None,
) -> UnitNullableRequiringPlayer:
    return UnitNullableRequiringPlayer(**udict)


def _player_from_dict(
    udict: dict[str, Any],
    definition: CreateTypeDefinition | None,
) -> Player:
    return Player(**udict)


def _unit_extended_player_from_dict(
    udict: dict[str, Any],
    definition: CreateTypeDefinition | None,
) -> UnitExtendedPlayer:
    return UnitExtendedPlayer(**udict)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.api_options import APIOptions

from ..collection_decimal_support_assets import (
    BASELINE_OBJ,
    S_OPTS_NO_DECS,
    S_OPTS_OK_DECS,
    WDECS_OBJ,
    is_decimal_super,
)
from ..conftest import DefaultAsyncCollection, DefaultCollection


class TestCollectionDecimalSupportIntegration:
    @pytest.mark.describe(
        "test of decimals not supported by default in collections, sync"
    )
    def test_decimalsupport_collections_defaultsettings_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        # write-and-read, baseline
        no_decimal_coll = sync_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=S_OPTS_NO_DECS,
            ),
        )
        no_decimal_coll.insert_one(BASELINE_OBJ)
        baseline_obj_2 = no_decimal_coll.find_one({"_id": BASELINE_OBJ["_id"]})
        assert baseline_obj_2 is not None
        assert BASELINE_OBJ == baseline_obj_2

        # the write should error for object with decimals
        with pytest.raises(TypeError):
            no_decimal_coll.insert_one(WDECS_OBJ)

    @pytest.mark.describe(
        "test of decimals not supported by default in collections, async"
    )
    async def test_decimalsupport_collections_defaultsettings_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        # write-and-read, baseline
        no_decimal_acoll = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=S_OPTS_NO_DECS,
            ),
        )
        await no_decimal_acoll.insert_one(BASELINE_OBJ)
        baseline_obj_2 = await no_decimal_acoll.find_one({"_id": BASELINE_OBJ["_id"]})
        assert baseline_obj_2 is not None
        assert BASELINE_OBJ == baseline_obj_2

        # the write should error for object with decimals
        with pytest.raises(TypeError):
            await no_decimal_acoll.insert_one(WDECS_OBJ)

    @pytest.mark.describe(
        "test of decimals supported in collections if set to do so, sync"
    )
    def test_decimalsupport_collections_decimalsettings_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        # write-and-read, baseline
        ok_decimal_coll = sync_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=S_OPTS_OK_DECS,
            ),
        )
        ok_decimal_coll.insert_one(BASELINE_OBJ)
        baseline_obj_2 = ok_decimal_coll.find_one({"_id": BASELINE_OBJ["_id"]})
        assert baseline_obj_2 is not None
        assert is_decimal_super(baseline_obj_2, BASELINE_OBJ)

        # write-and-read, with decimals
        ok_decimal_coll.insert_one(WDECS_OBJ)
        wdecs_2 = ok_decimal_coll.find_one({"_id": WDECS_OBJ["_id"]})
        assert wdecs_2 is not None
        assert is_decimal_super(wdecs_2, WDECS_OBJ)

    @pytest.mark.describe(
        "test of decimals supported in collections if set to do so, async"
    )
    async def test_decimalsupport_collections_decimalsettings_async(
        self,
        async_empty_collection: DefaultAsyncCollection,
    ) -> None:
        # write-and-read, baseline
        ok_decimal_acoll = async_empty_collection.with_options(
            api_options=APIOptions(
                serdes_options=S_OPTS_OK_DECS,
            ),
        )
        await ok_decimal_acoll.insert_one(BASELINE_OBJ)
        baseline_obj_2 = await ok_decimal_acoll.find_one({"_id": BASELINE_OBJ["_id"]})
        assert baseline_obj_2 is not None
        assert is_decimal_super(baseline_obj_2, BASELINE_OBJ)

        # write-and-read, with decimals
        await ok_decimal_acoll.insert_one(WDECS_OBJ)
        wdecs_2 = await ok_decimal_acoll.find_one({"_id": WDECS_OBJ["_id"]})
        assert wdecs_2 is not None
        assert is_decimal_super(wdecs_2, WDECS_OBJ)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bottleneck entrypoint for reading os.environ and exposing its contents as
(normalized) regular variables.
Except for the vectorize information, which for the time being passes as os.environ.
"""

from __future__ import annotations

import os
import time

from testcontainers.compose import DockerCompose

from astrapy.authentication import (
    StaticTokenProvider,
    TokenProvider,
    UsernamePasswordTokenProvider,
)
from astrapy.settings.defaults import DEFAULT_ASTRA_DB_KEYSPACE

DOCKER_COMPOSE_SLEEP_TIME_SECONDS = 20
DEFAULT_SECONDARY_KEYSPACE = "secondary_keyspace"

base_dir = os.path.abspath(os.path.dirname(__file__))
docker_compose_filepath = os.path.join(base_dir, "hcd_compose")


IS_ASTRA_DB: bool
DOCKER_COMPOSE_LOCAL_DATA_API: bool
SECONDARY_KEYSPACE: str
ASTRA_DB_API_ENDPOINT: str | None = None
ASTRA_DB_APPLICATION_TOKEN: str | None = None
ASTRA_DB_KEYSPACE: str | None = None
LOCAL_DATA_API_USERNAME: str | None = None
LOCAL_DATA_API_PASSWORD: str | None = None
LOCAL_DATA_API_ENDPOINT: str | None = None
LOCAL_CASSANDRA_CONTACT_POINT: str | None = None
LOCAL_CASSANDRA_PORT: str | None = None
LOCAL_DATA_API_KEYSPACE: str = DEFAULT_ASTRA_DB_KEYSPACE
RUN_SHARED_SECRET_VECTORIZE_TESTS: bool = True
USE_RERANKER_API_KEY_HEADER: bool = False

ASTRA_DB_TOKEN_PROVIDER: TokenProvider | None = None
LOCAL_DATA_API_TOKEN_PROVIDER: TokenProvider | None = None


def extended_booleanize_env(env_var_name: str, default: bool = False) -> bool:
    """
    Extended booleanize for environment variables.
    Accepts also "1"/"0", "yes"/"no", "true"/"false" (case insensitive).
    If the variable is not set, returns the default value.
    """
    value = os.environ.get(env_var_name)
    if value is None or value == "":
        return default
    # test if any integer:
    try:
        int_value = int(value)
        return int_value != 0
    except ValueError:
        pass
    value_lower = value.lower()
    if value_lower in ("yes", "true", "y", "t"):
        return True
    elif value_lower in ("no", "false", "n", "f"):
        return False
    else:
        raise ValueError(
            f"Invalid value for {env_var_name}: {value}. Expected an integer "
            "or one of 'yes', 'no', 'true', 'false', 'y', 'n', 't', 'f' (case insensitive)."
        )


# basic settings about which DB/API to use
if "LOCAL_DATA_API_ENDPOINT" in os.environ and os.environ["LOCAL_DATA_API_ENDPOINT"]:
    IS_ASTRA_DB = False
    DOCKER_COMPOSE_LOCAL_DATA_API = False
    LOCAL_DATA_API_USERNAME = os.environ.get("LOCAL_DATA_API_USERNAME")
    LOCAL_DATA_API_PASSWORD = os.environ.get("LOCAL_DATA_API_PASSWORD")
    LOCAL_DATA_API_ENDPOINT = os.environ["LOCAL_DATA_API_ENDPOINT"]
    LOCAL_CASSANDRA_PORT = os.environ.get("LOCAL_CASSANDRA_PORT")
    LOCAL_CASSANDRA_CONTACT_POINT = os.environ["LOCAL_CASSANDRA_CONTACT_POINT"]
    LOCAL_DATA_API_KEYSPACE = os.environ.get(
        "LOCAL_DATA_API_KEYSPACE", DEFAULT_ASTRA_DB_KEYSPACE
    )
    # no reason not to use it
    SECONDARY_KEYSPACE = os.environ.get(
        "LOCAL_DATA_API_SECONDARY_KEYSPACE", DEFAULT_SECONDARY_KEYSPACE
    )
elif "DOCKER_COMPOSE_LOCAL_DATA_API" in os.environ:
    IS_ASTRA_DB = False
    DOCKER_COMPOSE_LOCAL_DATA_API = True
    LOCAL_DATA_API_USERNAME = "cassandra"
    LOCAL_DATA_API_PASSWORD = "cassandra"
    LOCAL_DATA_API_ENDPOINT = "http://localhost:8181"
    LOCAL_CASSANDRA_CONTACT_POINT = "127.0.0.1"
    LOCAL_DATA_API_KEYSPACE = os.environ.get(
        "LOCAL_DATA_API_KEYSPACE", DEFAULT_ASTRA_DB_KEYSPACE
    )
    # no reason not to use it
    SECONDARY_KEYSPACE = os.environ.get(
        "LOCAL_DATA_API_SECONDARY_KEYSPACE", DEFAULT_SECONDARY_KEYSPACE
    )
elif "ASTRA_DB_API_ENDPOINT" in os.environ and os.environ["ASTRA_DB_API_ENDPOINT"]:
    IS_ASTRA_DB = True
    DOCKER_COMPOSE_LOCAL_DATA_API = False
    SECONDARY_KEYSPACE = os.environ.get(
        "ASTRA_DB_SECONDARY_KEYSPACE", DEFAULT_SECONDARY_KEYSPACE
    )
    ASTRA_DB_API_ENDPOINT = os.environ["ASTRA_DB_API_ENDPOINT"]
    ASTRA_DB_APPLICATION_TOKEN = os.environ["ASTRA_DB_APPLICATION_TOKEN"]
    ASTRA_DB_KEYSPACE = os.environ.get("ASTRA_DB_KEYSPACE", DEFAULT_ASTRA_DB_KEYSPACE)
else:
    raise ValueError("No credentials.")

RUN_SHARED_SECRET_VECTORIZE_TESTS = extended_booleanize_env(
    "RUN_SHARED_SECRET_VECTORIZE_TESTS", default=True
)

if os.environ.get("HEADER_RERANKING_API_KEY_NVIDIA") is not None:
    USE_RERANKER_API_KEY_HEADER = True

# token provider setup
if IS_ASTRA_DB:
    ASTRA_DB_TOKEN_PROVIDER = StaticTokenProvider(ASTRA_DB_APPLICATION_TOKEN)
else:
    # there must be a user/pwd pair
    if LOCAL_DATA_API_USERNAME and LOCAL_DATA_API_PASSWORD:
        LOCAL_DATA_API_TOKEN_PROVIDER = UsernamePasswordTokenProvider(
            username=LOCAL_DATA_API_USERNAME,
            password=LOCAL_DATA_API_PASSWORD,
        )
    else:
        raise ValueError("No full authentication data for local Data API")


# Ensure docker compose, if needed, is started and ready before anything else
# (especially querying the findEmbeddingProviders)
# if "DOCKER_COMPOSE_LOCAL_DATA_API", must spin the whole environment:
# (it is started and not cleaned up at the moment: manual cleanup if needed)
is_docker_compose_started = False
if DOCKER_COMPOSE_LOCAL_DATA_API:
    if not is_docker_compose_started:
        """
        Note: this is a trick to invoke `docker compose` as opposed to `docker-compose`
        while using testcontainers < 4.
        This trick is only reliable with testcontainers >= 3.1, < 4.

        The reason is that `docker-compose` is not included in the `ubuntu-latest`
        Github runner starting August 2024 (in favour of `docker compose`).

        - More modern testcontainers would require python>=3.9.
        - Aliasing 'docker-compose' in the CI container is (more) brittle.
        - Pip installing a modern testcontainers in the CI yaml is 'inelegant'.
            (and would require a try/except here with different inits to keep compat.)

        So we ended up with this trick, which redefines the executable to invoke
        for the DockerCompose class of testcontainers.
        """

        class RedefineCommandDockerCompose(DockerCompose):
            def docker_compose_command(self) -> list[str]:
                docker_compose_cmd = [
                    os.environ.get("DOCKER_COMMAND_NAME", "docker"),
                    "compose",
                ]
                for file in self.compose_file_names:
                    docker_compose_cmd += ["-f", file]
                if self.env_file:
                    docker_compose_cmd += ["--env-file", self.env_file]
                return docker_compose_cmd

        compose = RedefineCommandDockerCompose(filepath=docker_compose_filepath)
        compose.start()
        time.sleep(DOCKER_COMPOSE_SLEEP_TIME_SECONDS)
        is_docker_compose_started = True


# admin-test values for Astra DB
ADMIN_ENV_LIST = ["prod", "dev", "test"]
ADMIN_ENV_VARIABLE_MAP = {
    admin_env: {
        "token": os.environ.get(
            f"{admin_env.upper()}_ADMIN_TEST_ASTRA_DB_APPLICATION_TOKEN"
        ),
        "provider": os.environ.get(f"{admin_env.upper()}_ADMIN_TEST_ASTRA_DB_PROVIDER"),
        "region": os.environ.get(f"{admin_env.upper()}_ADMIN_TEST_ASTRA_DB_REGION"),
    }
    for admin_env in ADMIN_ENV_LIST
}

# misc variables
HEADER_EMBEDDING_API_KEY_OPENAI = os.environ.get("HEADER_EMBEDDING_API_KEY_OPENAI")
HEADER_RERANKING_API_KEY_NVIDIA = os.environ.get("HEADER_RERANKING_API_KEY_NVIDIA")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from ..conftest import (
    ADMIN_ENV_LIST,
    ADMIN_ENV_VARIABLE_MAP,
    CQL_AVAILABLE,
    HEADER_EMBEDDING_API_KEY_OPENAI,
    IS_ASTRA_DB,
    RUN_SHARED_SECRET_VECTORIZE_TESTS,
    SECONDARY_KEYSPACE,
    TEST_COLLECTION_NAME,
    USE_RERANKER_API_KEY_HEADER,
    VECTORIZE_TEXTS,
    DataAPICredentials,
    DataAPICredentialsInfo,
    DefaultAsyncCollection,
    DefaultAsyncTable,
    DefaultCollection,
    DefaultTable,
    _repaint_NaNs,
    _typify_tuple,
    async_fail_if_not_removed,
    clean_nulls_from_dict,
    sync_fail_if_not_removed,
)
from ..table_structure_assets import dict_equal_same_class
from ..table_udt_assets import (
    EXTENDED_PLAYER_TYPE_NAME,
    PLAYER_TYPE_NAME,
    THE_BYTES,
    THE_DATETIME,
    THE_TIMESTAMP,
    THE_TIMEZONE,
    ExtendedPlayer,
    NullablePlayer,
    Player,
    _extended_player_from_dict,
    _extended_player_serializer,
    _nullable_player_from_dict,
    _nullable_player_serializer,
    _player_from_dict,
    _player_serializer,
)

__all__ = [
    "DataAPICredentials",
    "DataAPICredentialsInfo",
    "async_fail_if_not_removed",
    "clean_nulls_from_dict",
    "sync_fail_if_not_removed",
    "HEADER_EMBEDDING_API_KEY_OPENAI",
    "IS_ASTRA_DB",
    "ADMIN_ENV_LIST",
    "ADMIN_ENV_VARIABLE_MAP",
    "CQL_AVAILABLE",
    "EXTENDED_PLAYER_TYPE_NAME",
    "PLAYER_TYPE_NAME",
    "RUN_SHARED_SECRET_VECTORIZE_TESTS",
    "SECONDARY_KEYSPACE",
    "TEST_COLLECTION_NAME",
    "THE_BYTES",
    "THE_DATETIME",
    "THE_TIMEZONE",
    "THE_TIMESTAMP",
    "USE_RERANKER_API_KEY_HEADER",
    "VECTORIZE_TEXTS",
    "dict_equal_same_class",
    "_extended_player_from_dict",
    "_extended_player_serializer",
    "_nullable_player_from_dict",
    "_nullable_player_serializer",
    "_player_from_dict",
    "_player_serializer",
    "_repaint_NaNs",
    "_typify_tuple",
    "DefaultCollection",
    "DefaultAsyncCollection",
    "DefaultAsyncTable",
    "DefaultTable",
    "ExtendedPlayer",
    "NullablePlayer",
    "Player",
]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main conftest for shared fixtures (if any).
"""

from __future__ import annotations

import functools
import warnings
from collections.abc import Awaitable, Iterable, Iterator
from typing import TYPE_CHECKING, Any, Callable, TypedDict

import pytest
from blockbuster import BlockBuster, blockbuster_ctx
from deprecation import UnsupportedWarning

if TYPE_CHECKING:
    from cassandra.cluster import Session
    from cassio.config import get_session_and_keyspace

import astrapy
from astrapy import AsyncDatabase, DataAPIClient, Database
from astrapy.admin import parse_api_endpoint
from astrapy.api_options import APIOptions, TimeoutOptions
from astrapy.authentication import TokenProvider
from astrapy.constants import Environment
from astrapy.settings.defaults import DEFAULT_ASTRA_DB_KEYSPACE, DEV_OPS_URL_ENV_MAP

from .preprocess_env import (
    ADMIN_ENV_LIST,
    ADMIN_ENV_VARIABLE_MAP,
    ASTRA_DB_API_ENDPOINT,
    ASTRA_DB_APPLICATION_TOKEN,
    ASTRA_DB_KEYSPACE,
    ASTRA_DB_TOKEN_PROVIDER,
    DOCKER_COMPOSE_LOCAL_DATA_API,
    HEADER_EMBEDDING_API_KEY_OPENAI,
    HEADER_RERANKING_API_KEY_NVIDIA,
    IS_ASTRA_DB,
    LOCAL_CASSANDRA_CONTACT_POINT,
    LOCAL_CASSANDRA_PORT,
    LOCAL_DATA_API_ENDPOINT,
    LOCAL_DATA_API_KEYSPACE,
    LOCAL_DATA_API_PASSWORD,
    LOCAL_DATA_API_TOKEN_PROVIDER,
    LOCAL_DATA_API_USERNAME,
    RUN_SHARED_SECRET_VECTORIZE_TESTS,
    SECONDARY_KEYSPACE,
    USE_RERANKER_API_KEY_HEADER,
)

CQL_AVAILABLE = False
try:
    from cassandra.auth import PlainTextAuthProvider
    from cassandra.cluster import Cluster, Session
    from cassio.config import get_session_and_keyspace

    CQL_AVAILABLE = True
except ImportError:
    pass
except Exception as exception:
    if exception.__class__.__name__ == "DependencyException":
        # this occurs on Python 3.13+ (as of Feb 2026) for libev/asyncore
        # TODO revise once newer cassandra-driver (>3.29) is out
        pass
    else:
        raise
    pass


@pytest.fixture(autouse=True)
def blockbuster() -> Iterator[BlockBuster]:
    with blockbuster_ctx("astrapy") as bb:
        # TODO: follow discussion in https://github.com/encode/httpx/discussions/3456
        bb.functions["os.stat"].can_block_in("httpx/_client.py", "_init_transport")
        yield bb


class DataAPICredentials(TypedDict):
    token: str | TokenProvider
    api_endpoint: str
    keyspace: str


class DataAPICredentialsInfo(TypedDict):
    environment: str
    region: str
    secondary_keyspace: str


def env_region_from_endpoint(api_endpoint: str) -> tuple[str, str]:
    parsed = parse_api_endpoint(api_endpoint)
    if parsed is not None:
        return (parsed.environment, parsed.region)
    else:
        return (Environment.OTHER, "no-region")


def database_id_from_endpoint(api_endpoint: str) -> str | None:
    parsed = parse_api_endpoint(api_endpoint)
    if parsed is not None:
        return parsed.database_id
    else:
        return None


def is_future_version(v_string: str) -> bool:
    def _my_int(st: str) -> int:
        rpos = st.find("rc")
        if rpos >= 0:
            return int(st[:rpos])
        return int(st)

    current_tuple = tuple(_my_int(pc) for pc in astrapy.__version__.split("."))
    v_tuple = tuple(_my_int(pc) for pc in v_string.split("."))
    return v_tuple > current_tuple


def async_fail_if_not_removed(
    method: Callable[..., Awaitable[Any]],
) -> Callable[..., Awaitable[Any]]:
    """
    Decorate a test async method to track removal of deprecated code.

    This is a customized+typed version of the deprecation package's
    `fail_if_not_removed` decorator (see), hastily put together to
    handle async test functions.

    See https://github.com/briancurtin/deprecation/issues/61 for reference.
    """

    @functools.wraps(method)
    async def test_inner(*args: Any, **kwargs: Any) -> Any:
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")
            rv = await method(*args, **kwargs)

        for warning in caught_warnings:
            if warning.category == UnsupportedWarning:
                raise AssertionError(
                    f"{method} uses a function that should be removed: {str(warning.message)}"
                )
        return rv

    return test_inner


def sync_fail_if_not_removed(method: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorate a test sync method to track removal of deprecated code.

    This is a typed version of the deprecation package's
    `fail_if_not_removed` decorator (see), with added minimal typing.
    """

    @functools.wraps(method)
    def test_inner(*args: Any, **kwargs: Any) -> Any:
        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")
            rv = method(*args, **kwargs)

        for warning in caught_warnings:
            if warning.category == UnsupportedWarning:
                raise AssertionError(
                    f"{method} uses a function that should be removed: {str(warning.message)}"
                )
        return rv

    return test_inner


def clean_nulls_from_dict(in_dict: dict[str, Any]) -> dict[str, Any]:
    def _cleand(_in: Any) -> Any:
        if isinstance(_in, list):
            return [_cleand(itm) for itm in _in]
        elif isinstance(_in, dict):
            return {k: _cleand(v) for k, v in _in.items() if v is not None}
        else:
            return _in

    return _cleand(in_dict)  # type: ignore[no-any-return]


@pytest.fixture(scope="session")
def data_api_credentials_kwargs() -> DataAPICredentials:
    api_creds: DataAPICredentials
    if IS_ASTRA_DB:
        if ASTRA_DB_API_ENDPOINT is None:
            raise ValueError("No endpoint data for local Data API")
        api_creds = {
            "token": ASTRA_DB_TOKEN_PROVIDER or "",
            "api_endpoint": ASTRA_DB_API_ENDPOINT or "",
            "keyspace": ASTRA_DB_KEYSPACE or DEFAULT_ASTRA_DB_KEYSPACE,
        }
    else:
        if LOCAL_DATA_API_ENDPOINT is None:
            raise ValueError("No endpoint data for local Data API")
        api_creds = {
            "token": LOCAL_DATA_API_TOKEN_PROVIDER or "",
            "api_endpoint": LOCAL_DATA_API_ENDPOINT or "",
            "keyspace": LOCAL_DATA_API_KEYSPACE or DEFAULT_ASTRA_DB_KEYSPACE,
        }

    # ensure keyspace(s) exist at this point
    # (we have to bypass the fixture hierarchy as the ..._info fixture
    # comes later, so this part instantiates and uses throwaway objects)
    _env, _ = env_region_from_endpoint(api_creds["api_endpoint"])
    _client = DataAPIClient(environment=_env)
    _database = _client.get_database(
        api_creds["api_endpoint"],
        token=api_creds["token"],
    )
    _database_admin = _database.get_database_admin()
    found_keyspaces = _database_admin.list_keyspaces()
    # This is an ugly way to reduce the risk of collision from concurrent
    # unit tests in CI/CD (which shows as "409 Conflict")
    if api_creds["keyspace"] not in found_keyspaces:
        _database_admin.create_keyspace(api_creds["keyspace"])
    found_keyspaces_2 = _database_admin.list_keyspaces()
    if SECONDARY_KEYSPACE not in found_keyspaces_2:
        _database_admin.create_keyspace(SECONDARY_KEYSPACE)
    # end of keyspace-ensuring block

    return api_creds


@pytest.fixture(scope="session")
def data_api_credentials_info(
    data_api_credentials_kwargs: DataAPICredentials,
) -> DataAPICredentialsInfo:
    api_endpoint = data_api_credentials_kwargs["api_endpoint"]
    env, reg = env_region_from_endpoint(api_endpoint)

    astra_db_cred_info: DataAPICredentialsInfo = {
        "environment": env,
        "region": reg,
        "secondary_keyspace": SECONDARY_KEYSPACE,
    }

    return astra_db_cred_info


@pytest.fixture(scope="session")
def client(
    data_api_credentials_info: DataAPICredentialsInfo,
) -> Iterable[DataAPIClient]:
    env = data_api_credentials_info["environment"]
    client = DataAPIClient(environment=env)
    yield client


@pytest.fixture(scope="session")
def sync_database(
    data_api_credentials_kwargs: DataAPICredentials,
    data_api_credentials_info: DataAPICredentialsInfo,
    client: DataAPIClient,
) -> Iterable[Database]:
    database = client.get_database(
        data_api_credentials_kwargs["api_endpoint"],
        token=data_api_credentials_kwargs["token"],
        keyspace=data_api_credentials_kwargs["keyspace"],
        spawn_api_options=APIOptions(
            timeout_options=TimeoutOptions(
                request_timeout_ms=20000,
                general_method_timeout_ms=60000,
                collection_admin_timeout_ms=90000,
                table_admin_timeout_ms=45000,
            ),
        ),
    )

    yield database


@pytest.fixture(scope="function")
def async_database(
    sync_database: Database,
) -> Iterable[AsyncDatabase]:
    yield sync_database.to_async()


@pytest.fixture(scope="session")
def cql_session(
    data_api_credentials_kwargs: DataAPICredentials,
) -> Iterable[Session]:
    if IS_ASTRA_DB:
        _env, _ = env_region_from_endpoint(data_api_credentials_kwargs["api_endpoint"])
        _db_id = database_id_from_endpoint(data_api_credentials_kwargs["api_endpoint"])

        if _db_id is None:
            raise ValueError("Could not extract database id for cql_session")

        _bundle_template = (
            f"{DEV_OPS_URL_ENV_MAP[_env]}/v2/databases/{{database_id}}/secureBundleURL"
        )

        _token: str
        if isinstance(data_api_credentials_kwargs["token"], TokenProvider):
            _token = data_api_credentials_kwargs["token"].get_token() or ""
        else:
            _token = data_api_credentials_kwargs["token"]

        if _token == "":
            raise ValueError("Token not found for cql_session")

        session, _ = get_session_and_keyspace(
            token=_token,
            database_id=_db_id,
            bundle_url_template=_bundle_template,
        )

        if session is None:
            raise ValueError("No CQL 'Session' was obtained")
        session.execute(f"USE {data_api_credentials_kwargs['keyspace']};")
        yield session
    else:
        if LOCAL_CASSANDRA_CONTACT_POINT is None:
            raise ValueError("No Cassandra contact point defined")

        auth_provider = PlainTextAuthProvider(
            username=LOCAL_DATA_API_USERNAME,
            password=LOCAL_DATA_API_PASSWORD,
        )
        additional_kwargs = {
            argk: argv
            for argk, argv in {
                "port": LOCAL_CASSANDRA_PORT,
            }.items()
            if argv is not None
        }
        cluster = Cluster(
            contact_points=[LOCAL_CASSANDRA_CONTACT_POINT],
            auth_provider=auth_provider,
            **additional_kwargs,
        )
        session = cluster.connect()
        session.execute(f"USE {data_api_credentials_kwargs['keyspace']};")
        yield session


__all__ = [
    "ASTRA_DB_API_ENDPOINT",
    "ASTRA_DB_APPLICATION_TOKEN",
    "ASTRA_DB_KEYSPACE",
    "CQL_AVAILABLE",
    "DOCKER_COMPOSE_LOCAL_DATA_API",
    "HEADER_EMBEDDING_API_KEY_OPENAI",
    "HEADER_RERANKING_API_KEY_NVIDIA",
    "IS_ASTRA_DB",
    "LOCAL_DATA_API_ENDPOINT",
    "LOCAL_DATA_API_KEYSPACE",
    "SECONDARY_KEYSPACE",
    "ADMIN_ENV_LIST",
    "ADMIN_ENV_VARIABLE_MAP",
    "ASTRA_DB_TOKEN_PROVIDER",
    "LOCAL_DATA_API_TOKEN_PROVIDER",
    "RUN_SHARED_SECRET_VECTORIZE_TESTS",
    "USE_RERANKER_API_KEY_HEADER",
]
